"""
Copyright 2014 Omer Gertel
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This code was originally developed by Omer Gertel, as a python port of the core portion of a
[Java Metrics library by Coda Hale](http://metrics.dropwizard.io/)

It was vendored into the inmanta source tree as the original was no longer maintained.
"""

import math
import time

from inmanta.vendor.pyformance import Clock


class ExpWeightedMovingAvg(object):
    """
    An exponentially-weighted moving average.
    """

    INTERVAL = 5.0  # seconds
    SECONDS_PER_MINUTE = 60.0

    def __init__(self, period: int, interval: float = INTERVAL, clock: Clock = time) -> None:
        """
        Create a new EWMA with a specific smoothing constant.

        :type period: C{int}
        :param period: the time in minutes it takes to reach a given significance level
        :type interval: C{int}
        :param interval: the expected tick interval, defaults to 5s
        """
        super(ExpWeightedMovingAvg, self).__init__()
        self.clock = clock
        self.uncounted = 0.0
        self.interval = interval
        self.rate: float = -1.0
        self.period = period * ExpWeightedMovingAvg.SECONDS_PER_MINUTE
        self.last_tick = self.clock.time()

    def get_rate(self) -> float:
        if self.clock.time() - self.last_tick >= self.interval:
            self.tick()
        if self.rate >= 0:
            return self.rate
        return 0

    def add(self, value: float) -> None:
        self.uncounted += value

    def tick(self) -> None:
        """
        Mark the passage of time and decay the current rate accordingly.
        """
        prev = self.last_tick
        now = self.clock.time()
        interval = now - prev
        if interval <= 0:
            return

        instant_rate = self.uncounted / interval
        self.uncounted = 0.0

        if self.rate >= 0:
            self.rate += self._alpha(interval) * (instant_rate - self.rate)
        else:
            self.rate = instant_rate

        self.last_tick = now

    def _alpha(self, interval: float) -> float:
        """
        Calculate the alpha based on the time since the last tick. This is
        necessary because a single threaded Python program loses precision
        under high load, so we can't assume a consistant I{EWMA._interval}.

        :type interval: C{float}
        :param interval: the interval we use to calculate the alpha
        """
        return 1 - math.exp(-interval / self.period)



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    """Change index on resource table to speed up query to get deployment progress in inmanta-lsm"""
    schema = """
-- Drop the old index
DROP INDEX IF EXISTS resource_environment_model_resource_type_idx;

-- Create the new index with the additional column
CREATE INDEX resource_environment_model_resource_type_idx
ON public.resource USING btree (environment, model, resource_type, resource_id_value);
    """
    await connection.execute(schema)



"""
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    """
    Split resource id into its composite value columns on the discovered resources table.
    """
    schema = r"""
    ALTER TABLE public.discoveredresource
        ADD COLUMN resource_type varchar,
        ADD COLUMN resource_id_value varchar,
        ADD COLUMN agent varchar
    ;
    UPDATE public.discoveredresource AS r
    SET (resource_type, agent, resource_id_value) = (
        SELECT rid.match[1], rid.match[3], rid.match[4]
        FROM (
            SELECT regexp_match(
                r.discovered_resource_id,
                '^(\w+(::\w+)*::[\w-]+)\[([^,]+),[^=]+=([^\]]*)]$'
            )
        ) AS rid(match)
    )
    ;
    ALTER TABLE public.discoveredresource
        ALTER COLUMN resource_type SET NOT NULL,
        ALTER COLUMN resource_id_value SET NOT NULL,
        ALTER COLUMN agent SET NOT NULL
    ;
    """
    await connection.execute(schema)



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import datetime
import logging
import os
import uuid
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, Optional

import inmanta.server.config as opt
from inmanta import config, const, data, protocol
from inmanta.agent import config as cfg
from inmanta.agent import executor, forking_executor
from inmanta.agent.reporting import collect_report
from inmanta.const import AGENT_SCHEDULER_ID
from inmanta.data.model import DataBaseReport, SchedulerStatusReport
from inmanta.deploy import scheduler
from inmanta.protocol import SessionEndpoint, methods, methods_v2
from inmanta.server.services.databaseservice import DatabaseMonitor
from inmanta.types import Apireturn
from inmanta.util import ensure_directory_exist, join_threadpools

LOGGER = logging.getLogger("inmanta.scheduler")


class Agent(SessionEndpoint):
    """
    This is the new scheduler, adapted to the agent protocol

    It serves a single endpoint that allows communications with the scheduler
    """

    def __init__(
        self,
        environment: Optional[uuid.UUID] = None,
    ):
        """
        :param environment: environment id
        """
        super().__init__(name="agent", timeout=cfg.server_timeout.get(), reconnect_delay=cfg.agent_reconnect_delay.get())

        self.thread_pool = ThreadPoolExecutor(1, thread_name_prefix="mainpool")
        self._storage = self.check_storage()

        if environment is None:
            environment = cfg.environment.get()
            if environment is None:
                raise Exception("The agent requires an environment to be set.")
        self.set_environment(environment)

        assert self._env_id is not None

        self.executor_manager: executor.ExecutorManager[executor.Executor] = self.create_executor_manager()
        self.scheduler = scheduler.ResourceScheduler(self._env_id, self.executor_manager, self._client)
        self.working = False

    async def start(self) -> None:
        # Make mypy happy
        assert data.Resource._connection_pool is not None
        self._db_monitor = DatabaseMonitor(
            data.Resource._connection_pool,
            opt.db_name.get(),
            opt.db_host.get(),
        )
        self._db_monitor.start()

        await super().start()

    def create_executor_manager(self) -> executor.ExecutorManager[executor.Executor]:
        assert self._env_id is not None
        return forking_executor.MPManager(
            self.thread_pool,
            self.sessionid,
            self._env_id,
            config.log_dir.get(),
            self._storage["executors"],
            LOGGER.level,
            cli_log=False,
        )

    async def stop(self) -> None:
        if self.working:
            await self.stop_working()
        if self._db_monitor:
            await self._db_monitor.stop()
        threadpools_to_join = [self.thread_pool]
        await self.executor_manager.join(threadpools_to_join, const.SHUTDOWN_GRACE_IOLOOP * 0.9)
        self.thread_pool.shutdown(wait=False)
        await join_threadpools(threadpools_to_join)
        await super().stop()

    async def start_connected(self) -> None:
        """
        Setup our single endpoint
        """
        await self.add_end_point_name(AGENT_SCHEDULER_ID)

    async def start_working(self) -> None:
        """Start working, once we have a session"""

        if self.working:
            return
        self.working = True
        await self.executor_manager.start()
        await self.scheduler.start()
        LOGGER.info("Scheduler started for environment %s", self.environment)

    async def stop_working(self, timeout: float = 0.0) -> None:
        """Stop working, connection lost"""
        if not self.working:
            return
        self.working = False
        await self.scheduler.stop()
        await self.executor_manager.stop()
        await self.executor_manager.join([], timeout=timeout)
        await self.scheduler.join()
        LOGGER.info("Scheduler stopped for environment %s", self.environment)

    @protocol.handle(methods_v2.remove_executor_venvs)
    async def remove_executor_venvs(self) -> None:
        """
        Remove all the venvs used by the executors of this agent.
        """
        try:
            await data.Notification(
                environment=self._env_id,
                created=datetime.datetime.now().astimezone(),
                title="Agent operations suspended",
                message="Agent operations are temporarily suspended because the user requested to remove the agent venvs.",
                severity=const.NotificationSeverity.info,
            ).insert()
            # Stop all deployments and stop all executors
            await self.scheduler.suspend_deployments(reason="removing all agent venvs")
            await self.executor_manager.stop_all_executors()
            # Remove venvs
            await self._remove_executor_venvs()
        except Exception as e:
            await data.Notification(
                environment=self._env_id,
                created=datetime.datetime.now().astimezone(),
                title="Agent venv removal failed",
                message=f"Failed to remove agent venvs: {e}",
                severity=const.NotificationSeverity.error,
            ).insert()
        else:
            await data.Notification(
                environment=self._env_id,
                created=datetime.datetime.now().astimezone(),
                title="Agent venv removal finished",
                message="The agent venvs were successfully removed. Resuming agent operations.",
                severity=const.NotificationSeverity.info,
            ).insert()
        finally:
            # Resume deployments again
            await self.scheduler.resume_deployments()

    async def _remove_executor_venvs(self) -> None:
        """
        This method was created to be able to monkeypatch it in testing.
        """
        environment_manager: executor.VirtualEnvironmentManager | None = self.executor_manager.get_environment_manager()
        if not environment_manager:
            raise Exception(
                "Calling the remove_executor_venvs endpoint while running against an ExecutorManager that doesn't have"
                " a VirtualEnvironmentManager. This can happen while running the test suite using"
                " the agent fixture. In that case all executors run in the same process as the server."
                " So there are no venvs to cleanup."
            )
        await environment_manager.remove_all_venvs()

    @protocol.handle(methods.set_state)
    async def set_state(self, agent: Optional[str], enabled: bool) -> Apireturn:
        if agent == AGENT_SCHEDULER_ID:
            if enabled:
                if self.working != enabled:
                    await self.start_working()
                else:
                    # Special case that the server considers us disconnected, but the Scheduler thinks we are still connected.
                    # In that case, the Scheduler may have missed some event, therefore, we need to refresh everything
                    # (Scheduler side) to make sure we are up to date when the server considers that the Scheduler
                    # is back online
                    await self.scheduler.read_version()
                    await self.scheduler.refresh_all_agent_states_from_db()
            else:
                # We want the request to not end in a 500 error:
                # if the scheduler is being shut down, it cannot respond to the request
                await self.stop_working(timeout=const.EXECUTOR_GRACE_HARD)

            return 200, "Scheduler has been notified!"
        else:
            if agent is None:
                await self.scheduler.refresh_all_agent_states_from_db()
                return 200, "All agents have been notified!"
            else:
                await self.scheduler.refresh_agent_state_from_db(name=agent)
                return 200, f"Agent `{agent}` has been notified!"

    async def on_reconnect(self) -> None:
        result = await self._client.get_state(tid=self._env_id, sid=self.sessionid, agent=AGENT_SCHEDULER_ID)
        if result.code == 200 and result.result is not None:
            state = result.result
            if "enabled" in state and isinstance(state["enabled"], bool):
                if state["enabled"]:
                    await self.start_working()
                else:
                    assert not self.working
            else:
                LOGGER.warning("Server reported invalid state %s" % (repr(state)))
        else:
            LOGGER.warning("could not get state from the server")

    async def on_disconnect(self) -> None:
        LOGGER.warning("Connection to server lost, stopping scheduler in environment %s", self.environment)
        await self.stop_working()

    @protocol.handle(methods.trigger, env="tid", agent="id")
    async def trigger_update(self, env: uuid.UUID, agent: None | str, incremental_deploy: bool) -> Apireturn:
        """
        Trigger an update for a specific agent, or for ALL agents in the environment when <agent> param is None.
        """
        if agent == const.AGENT_SCHEDULER_ID:
            agent = None

        if agent is None:
            agent_id = "All agents"
        else:
            agent_id = f"Agent {agent}"

        if incremental_deploy:
            LOGGER.info("%s got a trigger to run deploy in environment %s", agent_id, env)
            await self.scheduler.deploy(reason="user requested a deploy", agent=agent)
        else:
            LOGGER.info("%s got a trigger to run repair in environment %s", agent_id, env)
            await self.scheduler.repair(reason="user requested a repair", agent=agent)
        return 200

    @protocol.handle(methods.trigger_read_version, env="tid", agent="id")
    async def read_version(self, env: uuid.UUID) -> Apireturn:
        """
        Send a notification to the scheduler that a new version has been released
        """
        assert env == self.environment
        await self.scheduler.read_version()
        return 200

    @protocol.handle(methods.do_dryrun, env="tid", dry_run_id="id")
    async def run_dryrun(self, env: uuid.UUID, dry_run_id: uuid.UUID, agent: str, version: int) -> Apireturn:
        """
        Run a dryrun of the given version

        Paused agents are silently ignored
        """
        assert env == self.environment
        assert agent == AGENT_SCHEDULER_ID
        LOGGER.info("Agent %s got a trigger to run dryrun %s for version %s in environment %s", agent, dry_run_id, version, env)

        await self.scheduler.dryrun(dry_run_id, version)
        return 200

    @protocol.handle(methods.get_parameter, env="tid")
    async def get_facts(self, env: uuid.UUID, agent: str, resource: dict[str, Any]) -> Apireturn:
        # FIXME: this api is very inefficient: it sends the entire resource, we only need the id now
        assert env == self.environment
        assert agent == AGENT_SCHEDULER_ID
        LOGGER.info("Agent %s got a trigger to run get_facts for resource %s in environment %s", agent, resource.get("id"), env)
        await self.scheduler.get_facts(resource)
        return 200

    @protocol.handle(methods.get_status)
    async def get_status(self) -> Apireturn:
        return 200, collect_report(self)

    @protocol.handle(methods_v2.trigger_get_status, env="tid")
    async def get_scheduler_resource_state(self, env: data.Environment) -> SchedulerStatusReport:
        assert env.id == self.environment
        report = await self.scheduler.get_resource_state()
        return report

    @protocol.handle(methods_v2.notify_timer_update, env="tid")
    async def notify_timer_update(self, env: data.Environment) -> None:
        assert env == self.environment
        await self.scheduler.load_timer_settings()

    @protocol.handle(methods_v2.get_db_status)
    async def get_db_status(self) -> DataBaseReport:
        if self._db_monitor is None:
            return DataBaseReport(
                connected=False,
                database="",
                host="",
                max_pool=0,
                open_connections=0,
                free_connections=0,
                pool_exhaustion_count=0,
            )
        return await self._db_monitor.get_status()

    def check_storage(self) -> dict[str, str]:
        """
        Check if the server storage is configured and ready to use. Ultimately, this is
        what the layout on disk will look like:

            /var/lib/inmanta/
                ├─ server
                    ├─ env_uuid
                        ├─ executors/
                        │   ├─ venvs/
                        │   │   ├─ venv_blueprint_hash_1/
                        │   │   ├─ venv_blueprint_hash_2/
                        │   │   ├─ ...
                        │   │
                        │   ├─ code/
                        │       ├─ executor_blueprint_hash_1/
                        │       ├─ executor_blueprint_hash_2/
                        │       ├─ ...
                        │
                        ├─ compiler/
                        │
                        ├─ scheduler.cfg

        """

        state_dir = cfg.state_dir.get()
        if not os.path.exists(state_dir):
            os.mkdir(state_dir)

        dir_map = {
            "executors": ensure_directory_exist(state_dir, "executors"),
        }
        return dir_map



"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import contextlib
import dataclasses
import datetime
import enum
import functools
import hashlib
import importlib.metadata
import inspect
import itertools
import json
import logging
import os
import pathlib
import socket
import threading
import time
import typing
import uuid
import warnings
from abc import ABC, abstractmethod
from asyncio import AbstractEventLoop, CancelledError, Lock, Task, ensure_future, gather
from collections import abc, defaultdict
from collections.abc import Awaitable, Coroutine, Iterable, Iterator
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from logging import Logger
from types import TracebackType
from typing import TYPE_CHECKING, BinaryIO, Callable, Generic, Mapping, Optional, Sequence, TypeVar, Union

import asyncpg
import click
import pydantic
from tornado import gen

import packaging
import packaging.requirements
import packaging.utils
import pydantic_core
from crontab import CronTab
from inmanta import COMPILER_VERSION, const, types
from inmanta.stable_api import stable_api
from inmanta.types import JsonType, PrimitiveTypes, ReturnTypes
from packaging.utils import NormalizedName

if TYPE_CHECKING:
    from inmanta.data.model import ResourceId

LOGGER = logging.getLogger(__name__)
SALT_SIZE = 16
HASH_ROUNDS = 100000

T = TypeVar("T")
S = TypeVar("S")


def get_compiler_version() -> str:
    return COMPILER_VERSION


def groupby(mylist: list[T], f: Callable[[T], S]) -> Iterator[tuple[S, Iterator[T]]]:
    return itertools.groupby(sorted(mylist, key=f), f)


def ensure_directory_exist(directory: str, *subdirs: str) -> str:
    directory = os.path.join(directory, *subdirs)
    if not os.path.exists(directory):
        os.mkdir(directory)
    return directory


def is_sub_dict(subdct: dict[PrimitiveTypes, PrimitiveTypes], dct: dict[PrimitiveTypes, PrimitiveTypes]) -> bool:
    return not any(True for k, v in subdct.items() if k not in dct or dct[k] != v)


def strtobool(val: str) -> bool:
    """Convert a string representation of truth to True or False.

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.

    This function is based on a function in the Python distutils package. Is is subject
    to the following license:

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError("invalid truth value %r" % (val,))


def hash_file(content: bytes) -> str:
    """
    Create a hash from the given content
    """
    sha1sum = hashlib.new("sha1")
    sha1sum.update(content)

    return sha1sum.hexdigest()


def hash_file_streaming(file_handle: BinaryIO) -> str:
    h = hashlib.new("sha1")
    while True:
        # Reading is buffered, so we can read smaller chunks.
        chunk = file_handle.read(h.block_size)
        if not chunk:
            break
        h.update(chunk)

    return h.hexdigest()


def is_call_ok(result: Union[int, tuple[int, JsonType]]) -> bool:
    if isinstance(result, tuple):
        if len(result) == 2:
            code, reply = result
        else:
            raise Exception("Handlers for method call can only return a status code and a reply")

    else:
        code = result

    return code == 200


def ensure_future_and_handle_exception(
    logger: Logger, msg: str, action: Coroutine[object, None, T], notify_done_callback: Callable[[Task[T]], None]
) -> Task[T]:
    """Fire off a coroutine from the ioloop thread and log exceptions to the logger with the message"""
    future: Task[T] = ensure_future(action)

    def handler(task: Task[T]) -> None:
        try:
            exc = task.exception()
            if exc is not None:
                logger.exception(msg, exc_info=exc)
        except CancelledError:
            pass
        finally:
            notify_done_callback(task)

    future.add_done_callback(handler)
    return future


# In this module we use Coroutine[object, None, T] instead of Awaitable[T]. The reason for this is:
# - We use the methods here to background coroutines, which generates tasks.
# - Future is a subclass of Awaitable
# - ensure_future which is used in a number of methods here, returns Task unless you pass it a future.
# By only allowing Coroutines, we can make the typing more strict and only consider tasks.

TaskMethod = Callable[[], Coroutine[object, None, object]]


@stable_api
class TaskSchedule(ABC):
    """
    Abstract base class for a task schedule specification. Offers methods to inspect when the task should be scheduled, relative
    to the current time. Stateless.
    """

    @abstractmethod
    def get_initial_delay(self) -> float:
        """
        Returns the number of seconds from now until this task should be scheduled for the first time.
        """

    @abstractmethod
    def get_next_delay(self) -> float:
        """
        Returns the number of seconds from now until this task should be scheduled again. Uses the current time as reference,
        i.e. assumes the task has already run at least once and a negligible amount of time has passed since the last run
        completed.
        """

    @abstractmethod
    def log(self, action: TaskMethod) -> None:
        """
        Log a message about the action being scheduled according to this schedule.
        """


@stable_api
@dataclass(frozen=True)
class ScheduledTask:
    action: TaskMethod
    schedule: TaskSchedule


@stable_api
@dataclass(frozen=True)
class IntervalSchedule(TaskSchedule):
    """
    Simple interval schedule for tasks.

    :param interval: The interval in seconds between execution of actions.
    :param initial_delay: Delay in seconds to the first execution. If not set, interval is used.
    """

    interval: float
    initial_delay: Optional[float] = None

    def get_initial_delay(self) -> float:
        return self.initial_delay if self.initial_delay is not None else self.interval

    def get_next_delay(self) -> float:
        return self.interval

    def log(self, action: TaskMethod) -> None:
        LOGGER.debug(
            "Scheduling action %s every %d seconds with initial delay %d", action, self.interval, self.get_initial_delay()
        )


@stable_api
@dataclass(frozen=True)
class CronSchedule(TaskSchedule):
    """
    Current-time based scheduler: interval is calculated dynamically based on cron specifier and current time. Cron schedule is
    always interpreted as UTC.
    """

    cron: str
    _crontab: CronTab = dataclasses.field(init=False, compare=False)

    def __post_init__(self) -> None:
        crontab: CronTab
        try:
            crontab = CronTab(self.cron)
        except ValueError as e:
            raise ValueError(f"'{self.cron}' is not a valid cron expression: {e}")
        # can not assign directly on frozen dataclass, see dataclass docs
        object.__setattr__(self, "_crontab", crontab)

    def get_initial_delay(self) -> float:
        # no special treatment for first execution
        return self.get_next_delay()

    def get_next_delay(self) -> float:
        # always interpret cron schedules as UTC
        now: datetime.datetime = datetime.datetime.now(datetime.timezone.utc)
        return self._crontab.next(now=now)

    def log(self, action: TaskMethod) -> None:
        LOGGER.debug("Scheduling action %s according to cron specifier '%s'", action, self.cron)


def is_coroutine(function: object) -> bool:
    return (
        inspect.iscoroutinefunction(function)
        or gen.is_coroutine_function(function)
        or isinstance(function, functools.partial)
        and is_coroutine(function.func)
    )


@stable_api
class Scheduler:
    """
    An event scheduler class. Identifies tasks based on an action and a schedule. Considers tasks with the same action and the
    same schedule to be the same. Callers that wish to be able to delete the tasks they add should make sure to use unique
    `call` functions.
    Assumes an event loop is already running on this thread.
    """

    def __init__(self, name: str) -> None:
        self.name = name
        self._scheduled: dict[ScheduledTask, asyncio.TimerHandle] = {}
        self._stopped = False
        # Keep track of all tasks that are currently executing to be
        # able to cancel them when the scheduler is stopped.
        self._executing_tasks: dict[TaskMethod, list[asyncio.Task[object]]] = defaultdict(list)
        # Keep track of tasks that should be awaited before the scheduler is stopped
        self._await_tasks: dict[TaskMethod, list[asyncio.Task[object]]] = defaultdict(list)

    def _add_to_executing_tasks(self, action: TaskMethod, task: asyncio.Task[object], cancel_on_stop: bool = True) -> None:
        """
        Add task that is currently executing to `self._executing_tasks`.
        """
        # requires: RequiresProvidesMapping = dataclasses.field(default_factory=RequiresProvidesMapping)
        # # types per agent keeps track of which resource types live on which agent by doing a reference count
        # # the dict is agent_name -> resource_type -> resource_count
        # types_per_agent: dict[str, dict["ResourceType", int]] = dataclasses.field(
        #     default_factory=lambda: defaultdict(lambda: defaultdict(lambda: 0))
        # )
        if action in self._executing_tasks and self._executing_tasks[action]:
            LOGGER.warning("Multiple instances of background task %s are executing simultaneously", action.__name__)
        self._executing_tasks[action].append(task)
        if not cancel_on_stop:
            self._await_tasks[action].append(task)

    def _notify_done(self, action: TaskMethod, task: asyncio.Task[object]) -> None:
        """
        Called by the callback function of executing task when the task has finished executing.
        """

        def remove_action_from_task_dict(task_dict: dict[TaskMethod, list[asyncio.Task[object]]]) -> None:
            if action in task_dict:
                try:
                    task_dict[action].remove(task)
                except ValueError:
                    pass

        for task_dict in [self._executing_tasks, self._await_tasks]:
            remove_action_from_task_dict(task_dict)

    @stable_api
    def add_action(
        self,
        action: TaskMethod,
        schedule: Union[TaskSchedule, int],  # int for backward compatibility,
        cancel_on_stop: bool = True,
        quiet_mode: bool = False,
    ) -> Optional[ScheduledTask]:
        """
        Add a new action

        :param action: A function to call periodically
        :param schedule: The schedule for this action
        :param cancel_on_stop: Cancel the task when the scheduler is stopped. If false, the coroutine will be awaited.
        :param quiet_mode: Set to true to disable logging the recurring  notification that the action is being called.
        Use this to avoid polluting the server log for very frequent actions.
        """
        assert is_coroutine(action)

        if self._stopped:
            LOGGER.warning("Scheduling action '%s', while scheduler is stopped", action.__name__)
            return None

        schedule_typed: TaskSchedule
        if isinstance(schedule, int):
            schedule_typed = IntervalSchedule(schedule)
        else:
            schedule_typed = schedule

        schedule_typed.log(action)

        task_spec: ScheduledTask = ScheduledTask(action, schedule_typed)
        if task_spec in self._scheduled:
            # start fresh to respect initial delay, if set
            self.remove(task_spec)

        def action_function() -> None:
            if not quiet_mode:
                LOGGER.info("Calling %s", action.__name__)
            if task_spec in self._scheduled:
                try:
                    task = ensure_future_and_handle_exception(
                        logger=LOGGER,
                        msg="Uncaught exception while executing scheduled action",
                        action=action(),
                        notify_done_callback=functools.partial(self._notify_done, action),
                    )
                    self._add_to_executing_tasks(action, task, cancel_on_stop)
                except Exception:
                    LOGGER.exception("Uncaught exception while executing scheduled action")
                finally:
                    # next iteration
                    ihandle = asyncio.get_running_loop().call_later(schedule_typed.get_next_delay(), action_function)
                    self._scheduled[task_spec] = ihandle

        handle: asyncio.TimerHandle = asyncio.get_running_loop().call_later(schedule_typed.get_initial_delay(), action_function)
        self._scheduled[task_spec] = handle
        return task_spec

    @stable_api
    def remove(self, task: ScheduledTask) -> None:
        """
        Remove a scheduled action
        """
        if task in self._scheduled:
            self._scheduled[task].cancel()
            del self._scheduled[task]

    @stable_api
    async def stop(self) -> None:
        """
        Stop the scheduler
        """
        self._stopped = True
        try:
            # remove can still run during stop. That is why we loop until we get a keyerror == the dict is empty
            while True:
                _, handle = self._scheduled.popitem()
                handle.cancel()
        except KeyError:
            pass

        # Cancel all tasks that are already executing
        for action, tasks in self._executing_tasks.items():
            for task in tasks:
                if task not in self._await_tasks[action]:
                    task.cancel()

        results = await gather(
            *[handle for handles in self._await_tasks.values() for handle in handles], return_exceptions=True
        )

        # Log any exception that happened during shutdown
        for result in results:
            if isinstance(result, CancelledError):
                # Ignore this, it is ok to leak a cancel here
                pass
            if isinstance(result, Exception):
                LOGGER.error("Exception during shutdown", exc_info=result)

    def __del__(self) -> None:
        if len(self._scheduled) > 0:
            warnings.warn("Deleting scheduler '%s' that has not been stopped properly." % self.name)


def get_free_tcp_port() -> str:
    """
    Semi safe method for getting a random port. This may contain a race condition.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as tcp:
        tcp.bind(("", 0))
        _addr, port = tcp.getsockname()
        return str(port)


def datetime_iso_format(timestamp: datetime.datetime, *, tz_aware: bool = True) -> str:
    """
    Returns a timestamp ISO string.


    :param timestamp: The timestamp to get the ISO string for.
    :param tz_aware: Whether to return timezone aware timestamps or naive, implicit UTC timestamp.
    """

    def convert_timestamp() -> datetime.datetime:
        if tz_aware:
            if timestamp.tzinfo:
                return timestamp
            return timestamp.replace(tzinfo=datetime.timezone.utc)

        if timestamp.tzinfo:
            return timestamp.astimezone(datetime.timezone.utc).replace(tzinfo=None)
        return timestamp

    return convert_timestamp().isoformat(timespec="microseconds")


def parse_timestamp(timestamp: str) -> datetime.datetime:
    """
    Parse a timestamp into a timezone aware object. Naive timestamps are assumed to be UTC.
    """
    try:
        return datetime.datetime.strptime(timestamp, const.TIME_ISOFMT + "%z")
    except ValueError:
        # interpret naive datetimes as UTC
        return datetime.datetime.strptime(timestamp, const.TIME_ISOFMT).replace(tzinfo=datetime.timezone.utc)


class JSONSerializable(ABC):
    """
    Instances of this class are JSON serializable. Concrete subclasses should implement json_serialization_step.
    """

    @abstractmethod
    def json_serialization_step(self) -> Union[ReturnTypes, "JSONSerializable"]:
        """
        Perform a step in the serialization process. Returns an other serializable object.
        The implementation should make sure each step progresses serialization so that successively
        calling JSONSerializable.default eventually resolves to a trivially serializable object.
        """
        raise NotImplementedError()


@stable_api
def internal_json_encoder(o: object) -> Union[ReturnTypes, "JSONSerializable"]:
    """
    A custom json encoder that knows how to encode other types commonly used by Inmanta from standard python libraries. This
    encoder is meant to be used internally.
    """
    if isinstance(o, datetime.datetime):
        # Internally, all naive datetime instances are assumed local. Returns ISO timestamp with explicit timezone offset.
        return _custom_json_encoder(o if o.tzinfo is not None else o.astimezone())

    return _custom_json_encoder(o)


@stable_api
def api_boundary_json_encoder(o: object, tz_aware: bool = True) -> Union[ReturnTypes, "JSONSerializable"]:
    """
    A custom json encoder that knows how to encode other types commonly used by Inmanta from standard python libraries. This
    encoder is meant to be used for API boundaries.
    :param tz_aware: Whether to serialize timestamps as timezone aware objects or as naive implicit UTC.
    """
    if isinstance(o, datetime.datetime):
        # Accross API boundaries, all naive datetime instances are assumed UTC.
        return datetime_iso_format(o, tz_aware=tz_aware)

    return _custom_json_encoder(o)


def _custom_json_encoder(o: object) -> Union[ReturnTypes, "JSONSerializable"]:
    """
    A custom json encoder that knows how to encode other types commonly used by Inmanta from standard python libraries
    """
    if isinstance(o, JSONSerializable):
        return o.json_serialization_step()

    if isinstance(o, (uuid.UUID, pydantic.AnyUrl, pydantic_core.Url)):
        return str(o)

    if isinstance(o, datetime.datetime):
        return o.isoformat(timespec="microseconds")

    if hasattr(o, "to_dict"):
        return o.to_dict()  # type: ignore

    if isinstance(o, enum.Enum):
        return o.name

    if isinstance(o, Exception):
        # Logs can push exceptions through RPC. Return a string representation.
        return str(o)

    from inmanta.data.model import BaseModel

    if isinstance(o, (BaseModel, pydantic.BaseModel)):
        return o.model_dump(by_alias=True)

    if dataclasses.is_dataclass(o) and not isinstance(o, type):
        return dataclasses.asdict(o)

    LOGGER.error("Unable to serialize %s", o)
    raise TypeError(repr(o) + " is not JSON serializable")


def add_future(future: Coroutine[object, None, T]) -> Task[T]:
    """
    Add a future to the ioloop to be handled, but do not require the result.
    """

    def handle_result(f: Task[T]) -> None:
        try:
            f.result()
        except Exception as e:
            LOGGER.exception("An exception occurred while handling a future: %s", str(e))

    task = ensure_future(future)
    task.add_done_callback(handle_result)
    return task


async def retry_limited(
    fun: Union[abc.Callable[..., bool], abc.Callable[..., abc.Awaitable[bool]]],
    timeout: float,
    interval: float = 0.1,
    *args: object,
    **kwargs: object,
) -> None:
    """
    This function makes use of the INMANTA_RETRY_LIMITED_MULTIPLIER env variable. If set, INMANTA_RETRY_LIMITED_MULTIPLIER
    serves as multiplier: The timeout given as argument becomes a 'soft limit' and the 'soft limit' multiplied by the
    multiplier (from the env var) becomes a 'hard limit'. If the hard limit is reached before the wait condition is fulfilled
    a Timeout exception is raised. If the wait condition is fulfilled before the hard limit is reached but after the soft
    limit is reached, a different Timeout exception is raised. if the Env var is not set, then the soft and hard limit are
    the same.
    """

    async def fun_wrapper() -> bool:
        if inspect.iscoroutinefunction(fun):
            return await fun(*args, **kwargs)
        else:
            return fun(*args, **kwargs)

    multiplier: int = int(os.environ.get("INMANTA_RETRY_LIMITED_MULTIPLIER", 1))
    if multiplier < 1:
        raise ValueError("value of INMANTA_RETRY_LIMITED_MULTIPLIER must be bigger or equal to 1.")
    hard_timeout = timeout * multiplier
    start = time.time()
    result = await fun_wrapper()
    while time.time() - start < hard_timeout and not result:
        await asyncio.sleep(interval)
        result = await fun_wrapper()
    if not result:
        raise asyncio.TimeoutError(f"Wait condition was not reached after hard limit of {hard_timeout} seconds")
    if time.time() - start > timeout:
        raise asyncio.TimeoutError(
            f"Wait condition was met after {time.time() - start} seconds, but soft limit was set to {timeout} seconds"
        )


class StoppedException(Exception):
    """This exception is raised when a background task is added to the taskhandler when it is shutting down."""


class TaskHandler(Generic[T]):
    """
    This class provides a method to add a background task based on a coroutine. When the coroutine ends, any exceptions
    are reported. If stop is invoked, all background tasks are cancelled.
    """

    def __init__(self) -> None:
        super().__init__()
        self._background_tasks: set[Task[T]] = set()
        self._await_tasks: set[Task[T]] = set()
        self._stopped = False

    def is_stopped(self) -> bool:
        return self._stopped

    def is_running(self) -> bool:
        return not self._stopped

    def add_background_task(self, future: Awaitable[T], cancel_on_stop: bool = True) -> Task[T]:
        """Add a background task to the event loop. When stop is called, the task is cancelled.

        :param future: The future or coroutine to run as background task.
        :param cancel_on_stop: Cancel the task when stop is called. If false, the coroutine is awaited.
        """
        if self._stopped:
            LOGGER.warning("Not adding background task because we are stopping.")
            raise StoppedException("A background tasks are not added to the event loop while stopping")

        task: Task[T] = ensure_future(future)

        def handle_result(task: Task[T]) -> None:
            try:
                task.result()
            except CancelledError:
                LOGGER.warning("Task %s was cancelled.", task)

            except Exception as e:
                LOGGER.exception("An exception occurred while handling a future: %s", str(e))
            finally:
                self._background_tasks.discard(task)
                self._await_tasks.discard(task)

        task.add_done_callback(handle_result)
        self._background_tasks.add(task)

        if not cancel_on_stop:
            self._await_tasks.add(task)

        return task

    async def stop(self) -> None:
        """Stop all background tasks by requesting a cancel"""
        self._stopped = True
        await gather(*self._await_tasks)
        self._background_tasks.difference_update(self._await_tasks)

        cancelled_tasks = []
        try:
            while True:
                task = self._background_tasks.pop()
                task.cancel()
                cancelled_tasks.append(task)
        except KeyError:
            pass

        await gather(*cancelled_tasks, return_exceptions=True)


class CycleException(Exception):
    def __init__(self, node: str) -> None:
        self.nodes = [node]
        self.done = False

    def add(self, node: str) -> None:
        if not self.done:
            if node not in self.nodes:
                self.nodes.insert(0, node)
            else:
                self.done = True


def stable_depth_first(nodes: list[str], edges: dict[str, list[str]]) -> list[str]:
    """Creates a linear sequence based on a set of "comes after" edges, same graph yields the same solution,
    independent of order given to this function"""
    nodes = sorted(nodes)
    edges = {k: sorted(v) for k, v in edges.items()}
    out = []

    def dfs(node: str, seen: set[str] = set()) -> None:
        if node in out:
            return
        if node in seen:
            raise CycleException(node)
        try:
            if node in edges:
                for edge in edges[node]:
                    dfs(edge, seen | {node})
            out.append(node)
        except CycleException as e:
            e.add(node)
            raise e

    while nodes:
        dfs(nodes.pop(0))

    return out


class NamedSubLock:
    def __init__(self, parent: "NamedLock", name: str) -> None:
        self.parent = parent
        self.name = name

    async def __aenter__(self) -> None:
        await self.parent.acquire(self.name)

    async def __aexit__(
        self, exc_type: Optional[type[BaseException]], exc_value: Optional[BaseException], traceback: Optional[TracebackType]
    ) -> None:
        await self.parent.release(self.name)


class NamedLock:
    """Create fine grained locks"""

    def __init__(self) -> None:
        self._master_lock: Lock = Lock()
        self._named_locks: dict[str, Lock] = {}
        self._named_locks_counters: dict[str, int] = {}

    def get(self, name: str) -> NamedSubLock:
        return NamedSubLock(self, name)

    async def acquire(self, name: str) -> None:
        async with self._master_lock:
            if name in self._named_locks:
                lock = self._named_locks[name]
                self._named_locks_counters[name] += 1
            else:
                lock = Lock()
                self._named_locks[name] = lock
                self._named_locks_counters[name] = 1
        await lock.acquire()

    async def release(self, name: str) -> None:
        async with self._master_lock:
            lock = self._named_locks[name]
            lock.release()
            self._named_locks_counters[name] -= 1
            # This relies on the internal mechanics of the lock
            if self._named_locks_counters[name] <= 0:
                del self._named_locks[name]
                del self._named_locks_counters[name]


class nullcontext(contextlib.nullcontext[T], contextlib.AbstractAsyncContextManager[T]):
    """
    nullcontext ported from Python 3.10 to support async
    """

    async def __aenter__(self) -> T:
        return self.enter_result

    async def __aexit__(self, *excinfo: object) -> None:
        pass


class FinallySet(contextlib.AbstractAsyncContextManager[asyncio.Event]):
    def __init__(self, event: asyncio.Event) -> None:
        self.event = event

    async def __aenter__(self) -> asyncio.Event:
        return self.event

    async def __aexit__(self, *exc_info: object) -> None:
        self.event.set()


async def join_threadpools(threadpools: list[ThreadPoolExecutor]) -> None:
    """
    Asynchronously join a set of threadpools

    idea borrowed from BaseEventLoop.shutdown_default_executor

    We implemented this method because:
    1. ThreadPoolExecutor.shutdown(wait=True)` is a blocking call, blocking the ioloop.
       This doesn't work because we often have back-and-forth between the ioloop and the thread
       due to our `ResourceHandler.run_sync` method.
    2.The python sdk has no support for async awaiting threadpool shutdown (except for the default pool)
    """

    loop = asyncio.get_running_loop()
    future = loop.create_future()

    def join() -> None:
        for threadpool in threadpools:
            try:
                threadpool.shutdown(wait=True)
            except Exception:
                LOGGER.exception("Exception during threadpool shutdown")
        loop.call_soon_threadsafe(future.set_result, None)

    thread = threading.Thread(target=join)
    thread.start()
    try:
        await future
    finally:
        thread.join()


def ensure_event_loop() -> asyncio.AbstractEventLoop:
    """
    Returns the event loop for this thread. Creates a new one if none exists yet and registers it with asyncio's active event
    loop policy. Does not ensure that the event loop is running.
    """
    try:
        # nothing needs to be done if this thread already has an event loop
        # known issue: asyncio offers no way to get the active event loop if it's not running. So we may be too eager
        #   in creating a new one.
        return asyncio.get_running_loop()
    except RuntimeError:
        # asyncio.set_event_loop sets the event loop for this thread only
        new_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(new_loop)
        return new_loop


def wait_sync[T](
    awaitable: Awaitable[T] | asyncio.Future[T],
    *,
    timeout: int = 120,
    ioloop: Optional[asyncio.AbstractEventLoop] = None,
) -> T:
    """
    Blocks on an async awaitable from a syncronous context. Must not be called from an async context.

    Must not be called from handlers, see Handler.run_sync for that.

    :param ioloop: await on an existing io loop, on another thread.
    If not given, the default loop (see `get_default_event_loop()`) is used.
    If it is not set an io loop is started on the current thread.
    """
    with_timeout: types.AsyncioCoroutine[T] = asyncio.wait_for(awaitable, timeout)

    if ioloop is None:
        # Fallback to ensure the agent doesn't leak ioloops via its threadpool
        ioloop = get_default_event_loop()

    if ioloop is not None:
        # run it on the given loop
        return asyncio.run_coroutine_threadsafe(with_timeout, ioloop).result()
    # no running loop given: create a loop for this thread if it doesn't exist already and run it
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        # no loop running for this thread. Use asyncio.run
        return asyncio.run(with_timeout)
    else:
        raise Exception("wait_sync can not be called from an async context")


class ExhaustedPoolWatcher:
    """
    This class keeps track of database pool exhaustion events and offers reporting capabilities.

    """

    def __init__(self, pool: asyncpg.pool.Pool) -> None:
        self._exhausted_pool_events_count: int = 0
        self._last_report: int = 0
        self._pool: asyncpg.pool.Pool = pool

    def report_and_reset(self, logger: logging.Logger) -> None:
        """
        Log how many exhausted pool events were recorded since the last time the counter
        was reset, if any, and reset the counter.
        """
        since_last = self._exhausted_pool_events_count - self._last_report
        if since_last > 0:
            logger.warning("Database pool was exhausted %d times in the past 24h.", since_last)
            self._last_report = self._exhausted_pool_events_count

    def check_for_pool_exhaustion(self) -> None:
        """
        Checks if the database pool is exhausted
        """
        pool_exhausted: bool = (self._pool.get_size() == self._pool.get_max_size()) and self._pool.get_idle_size() == 0
        if pool_exhausted:
            self._exhausted_pool_events_count += 1


def remove_comment_part_from_specifier(to_clean: str) -> str:
    """
    Remove the comment part of a requirement specifier

    :param to_clean: The requirement specifier to clean
    :return: A cleaned requirement specifier
    """
    # Refer to PEP 508. A requirement could contain a hashtag
    to_clean = to_clean.strip()
    drop_comment, _, _ = to_clean.partition(" #")
    # We make sure whitespaces are not counted in the length of this string, e.g. "        #"
    drop_comment = drop_comment.strip()
    return drop_comment


if typing.TYPE_CHECKING:

    class CanonicalRequirement(packaging.requirements.Requirement):
        name: NormalizedName

        def __init__(self, requirement: packaging.requirements.Requirement) -> None:
            raise Exception("Typing dummy, should never be seen")

else:
    CanonicalRequirement = typing.NewType("CanonicalRequirement", packaging.requirements.Requirement)
    """
    A CanonicalRequirement is a packaging.requirements.Requirement except that the name of this Requirement is canonicalized,
    which allows us to compare names without dealing afterwards with the format of these requirements.
    """


def parse_requirement(requirement: str) -> CanonicalRequirement:
    """
    Parse the given requirement string into a requirement object with a canonicalized name, meaning that we are sure that
    every CanonicalRequirement will follow the same convention regarding the name. This will allow us compare requirements.
    This function supposes to receive an actual requirement.

    :param requirement: The requirement's name
    :return: A new requirement instance
    """
    # We canonicalize the name of the requirement to be able to compare requirements and check if the requirement is
    # already installed
    # The following line could cause issue because we are not supposed to modify fields of an existing instance
    # The version of packaging is constrained to ensure this can not cause problems in production.
    requirement_instance = packaging.requirements.Requirement(requirement_string=requirement)
    requirement_instance.name = packaging.utils.canonicalize_name(requirement_instance.name)
    canonical_requirement_instance = CanonicalRequirement(requirement_instance)
    return canonical_requirement_instance


def parse_requirements(requirements: Sequence[str]) -> list[CanonicalRequirement]:
    """
    Parse the given requirements (sequence of strings) into requirement objects with a canonicalized name, meaning that we
    are sure that every CanonicalRequirement will follow the same convention regarding the name. This will allow us compare
    requirements. This function supposes to receive actual requirements. Commented strings will not be handled and result in
    a ValueError

    :param requirements: The names of the different requirements
    :return: list[CanonicalRequirement]
    """
    return [parse_requirement(requirement=e) for e in requirements]


def parse_requirements_from_file(file_path: pathlib.Path) -> list[CanonicalRequirement]:
    """
    Parse the given requirements (line by line) from a file into requirement objects with a canonicalized name, meaning that we
    are sure that every CanonicalRequirement will follow the same convention regarding the name. This will allow us compare
    requirements.

    :param file_path: The path to the read the requirements from
    :return: list[CanonicalRequirement]
    """
    if not file_path.exists():
        raise RuntimeError(f"The provided path does not exist: `{file_path}`!")

    with open(file_path) as f:
        file_contents: list[str] = f.readlines()
        requirements = [
            parse_requirement(remove_comment_part_from_specifier(line))
            for line in file_contents
            if (stripped := line.lstrip()) and not stripped.startswith("#")  # preprocessing
        ]

    return requirements


# Retaken from the `click-plugins` repo which is now unmaintained
def click_group_with_plugins(plugins: Iterable[importlib.metadata.EntryPoint]) -> Callable[[click.Group], click.Group]:
    """
    A decorator to register external CLI commands to an instance of `click.Group()`.

    :param plugins: An iterable producing one `pkg_resources.EntryPoint()` per iteration
    :return: The provided click group with the new commands
    """

    def decorator(group: click.Group) -> click.Group:
        for entry_point in plugins:
            try:
                group.add_command(entry_point.load())
            except Exception as e:
                # Catch this so a busted plugin doesn't take down the CLI.
                # Handled by registering a dummy command that does nothing
                # other than explain the error.
                def print_error(error: Exception) -> None:
                    click.echo(f"Error: could not load this plugin for the following reason: {error}")

                new_print_error = functools.partial(print_error, e)
                group.add_command(click.Command(name=entry_point.name, callback=new_print_error))

        return group

    return decorator


def make_attribute_hash(resource_id: "ResourceId", attributes: Mapping[str, object]) -> str:
    """
    This method returns the attribute hash for the attributes of the given resource.
    """
    from inmanta.protocol.common import custom_json_encoder

    character = json.dumps(
        {k: v for k, v in attributes.items() if k not in ["requires", "provides", "version"]},
        default=custom_json_encoder,
        sort_keys=True,  # sort the keys for stable hashes when using dicts, see #5306
    )
    m = hashlib.md5()
    m.update(resource_id.encode("utf-8"))
    m.update(character.encode("utf-8"))
    return m.hexdigest()


def get_pkg_name_and_version(path_distribution_pkg: str) -> tuple[str, str]:
    """
    Returns a tuple that holds the name and version number of the given
    distribution package. This method is compatible with both wheels and sdist packages.

    Sdist file format: `{name}-{version}.tar.gz`
    Wheel file format: `{distribution}-{version}(-{build tag})?-{python tag}-{abi tag}-{platform tag}.whl`
    """
    filename = os.path.basename(path_distribution_pkg)
    if filename.endswith(".tar.gz"):
        filename = filename.removesuffix(".tar.gz")
    pkg_name, version = filename.split("-")[0:2]
    normalized_pkg_name = parse_requirement(pkg_name.removeprefix("inmanta-module-")).name
    return normalized_pkg_name, version


def get_module_name(path_distribution_pkg: str) -> str:
    """
    Returns the name of the Inmanta module that belongs to the given python package.
    """
    filename = os.path.basename(path_distribution_pkg)
    pkg_name: str = filename.split("-", maxsplit=1)[0]
    return pkg_name.removeprefix("inmanta_module_")


default_event_loop: AbstractEventLoop | None = None


def set_default_event_loop(eventloop: AbstractEventLoop | None) -> None:
    global default_event_loop
    default_event_loop = eventloop


@stable_api
def get_default_event_loop() -> AbstractEventLoop | None:
    """
    Returns the default event loop.

    This is intended to be used by other threads, that want to run code on the eventloop of the main thread.

    It should be used to prevent leaking eventloops when using threadpools.
    The main use case is the `Client` class

    Only thread safe methods of the eventloop should be used.

    If an event loop is returned it will be running.
    """
    return default_event_loop



"""
Copyright 2014 Omer Gertel
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This code was originally developed by Omer Gertel, as a python port of the core portion of a
[Java Metrics library by Coda Hale](http://metrics.dropwizard.io/)

It was vendored into the inmanta source tree as the original was no longer maintained.
"""

from typing import Callable, Union


class Gauge[T: Union[int, float]]:
    """
    A base class for reading of a particular.

    For example, to instrument a queue depth:

    class QueueLengthGaguge(Gauge):
        def __init__(self, queue):
            super(QueueGaguge, self).__init__()
            self.queue = queue

        def get_value(self):
            return len(self.queue)

    """

    def get_value(self) -> T:
        "A subclass of Gauge should implement this method"
        raise NotImplementedError()


class CallbackGauge[T: Union[int, float]](Gauge[T]):
    """
    A Gauge reading for a given callback
    """

    def __init__(self, callback: Callable[[], T]) -> None:
        "constructor expects a callable"
        super(CallbackGauge, self).__init__()
        self.callback = callback

    def get_value(self) -> T:
        "returns the result of callback which is executed each time"
        return self.callback()


class SimpleGauge[T: Union[int, float]](Gauge[T]):
    """
    A gauge which holds values with simple getter- and setter-interface
    """

    def __init__(self, value: T) -> None:
        "constructor accepts initial value"
        super(SimpleGauge, self).__init__()
        self._value = value

    def get_value(self) -> T:
        "getter returns current value"
        return self._value

    def set_value(self, value: T) -> None:
        "setter changes current value"
        # XXX: add locking?
        self._value = value


type AnyGauge = Gauge[float | int]



"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    """
    Add the indexes to make the joins done by a cascading delete perform well.
    """
    schema = """
        -- Foreign key resourceaction_resource(resource_action_id) --> resource(action_id)
        CREATE INDEX IF NOT EXISTS resourceaction_resource_resource_action_id_index
            ON resourceaction_resource(resource_action_id);

        -- Required by query done in the ConfigurationModel.delete_cascade() method that deletes
        -- facts for resources that don't exist anymore.
        CREATE INDEX parameter_environment_resource_id_index ON parameter(environment, resource_id);
        CREATE INDEX resource_resource_id_index ON resource(resource_id);

        -- Foreign key Compile(substitute_compile_id) --> compile(id)
        CREATE INDEX compile_substitute_compile_id_index ON compile(substitute_compile_id);

        -- Drop partial index on agent(id_primary) and replace by a full index.
        -- Required for foreign key agent(id_primary) --> agentinstance(id)
        DROP INDEX agent_id_primary_index;
        CREATE INDEX agent_id_primary_index ON agent(id_primary);
    """
    await connection.execute(schema)



"""
Copyright 2021 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import enum
import importlib.metadata
import importlib.util
import json
import logging
import os
import re
import site
import subprocess
import sys
import tempfile
import typing
import venv
from collections import abc
from collections.abc import Iterator, Mapping, Sequence
from dataclasses import dataclass
from functools import reduce
from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from importlib.metadata import Distribution, distribution, distributions
from itertools import chain
from subprocess import CalledProcessError
from textwrap import indent
from typing import Callable, Optional, Tuple, TypeVar

import inmanta.util
import packaging.requirements
import packaging.utils
import packaging.version
from inmanta import const
from inmanta.ast import CompilerException
from inmanta.data.model import LEGACY_PIP_DEFAULT, PipConfig
from inmanta.server.bootloader import InmantaBootloader
from inmanta.stable_api import stable_api
from inmanta.util import parse_requirement, strtobool
from packaging.utils import NormalizedName, canonicalize_name

LOGGER = logging.getLogger(__name__)
LOGGER_PIP = logging.getLogger("inmanta.pip")  # Use this logger to log pip commands or data related to pip commands.

if typing.TYPE_CHECKING:
    from _typeshed.importlib import MetaPathFinderProtocol, PathEntryFinderProtocol


class PackageNotFound(Exception):
    pass


class PipInstallError(Exception):
    pass


@dataclass(eq=True, frozen=True)
class VersionConflict:
    """
    Represents a version conflict that exists in a Python environment.

    :param requirement: The requirement that is unsatisfied.
    :param installed_version: The version that is currently installed. None if the package is not installed.
    :param owner: The package from which the constraint originates
    """

    requirement: inmanta.util.CanonicalRequirement
    installed_version: Optional[packaging.version.Version] = None
    owner: Optional[str] = None

    def __str__(self) -> str:
        owner = ""
        if self.owner:
            # Cfr pip
            # Requirement already satisfied: certifi>=2017.4.17 in /[...]/site-packages
            # (from requests>=2.23.0->cookiecutter<3,>=1->inmanta-core==7.0.0) (2022.6.15)
            owner = f" (from {self.owner})"
        if self.installed_version:
            return (
                f"Incompatibility between constraint {self.requirement} and installed version {self.installed_version}{owner}"
            )
        else:
            return f"Constraint {self.requirement} is not installed{owner}"


class ConflictingRequirements(CompilerException):
    """
    Conflict reporting

    Can be used in two ways:
    - if we don't know the exact conflicts (detected by e.g. pip), the messages is used
    - if we have detailed conflict info, the message is derived from it

    """

    def __init__(self, message: str, conflicts: Optional[set[VersionConflict]] = None):
        CompilerException.__init__(self, msg=message)
        self.conflicts = conflicts

    def get_message(self) -> str:
        # The message has three potential parts
        # First the advices, derived from the conflicts, if present
        # Then the message, if present
        # Then the individual conflicts, if present
        out = []

        advices = self.get_advice()
        if advices:
            out.append(advices)

        if self.msg:
            out.append(self.msg)

        conflicts = self.get_conflicts_string()
        if conflicts:
            out.append(conflicts)

        return "\n".join(out)

    def get_conflicts_string(self) -> Optional[str]:
        if not self.conflicts:
            return None
        msg = ""
        for current_conflict in sorted(self.conflicts, key=lambda x: x.requirement.name):
            msg += f"\n\t* {current_conflict}"
        return msg

    def has_missing(self) -> bool:
        """Does the set of conflicts contain any missing dependency?"""
        if not self.conflicts:
            return False
        return any(conflict.installed_version is None for conflict in self.conflicts)

    def get_advice(self) -> Optional[str]:
        """
        Derive an end-user centric message from the conflicts
        """
        if self.conflicts is None:
            return None
        if self.has_missing():
            return "Not all required python packages are installed run 'inmanta project install' to resolve this"
        else:
            return (
                "A dependency conflict exists, this is either because some modules are stale, incompatible "
                "or because pip can not find a correct combination of packages. To resolve this, "
                "first try `inmanta project update` to ensure no modules are stale. "
                "Second, try adding additional constraints to the requirements.txt file of "
                "the inmanta project to help pip resolve this problem. After every change, run `inmanta project update`"
            )


req_list = TypeVar("req_list", Sequence[str], Sequence[inmanta.util.CanonicalRequirement])


class PythonWorkingSet:
    @classmethod
    def _get_as_requirements_type(cls, requirements: req_list) -> Sequence[inmanta.util.CanonicalRequirement]:
        """
        Convert requirements from Union[Sequence[str], Sequence[Requirement]] to Sequence[Requirement]
        """
        if isinstance(requirements[0], str):
            return inmanta.util.parse_requirements(requirements)
        else:
            return requirements

    @classmethod
    def are_installed(cls, requirements: req_list) -> bool:
        """
        Return True iff the given requirements are installed in this workingset.
        """
        if not requirements:
            return True
        installed_packages: dict[NormalizedName, packaging.version.Version] = cls.get_packages_in_working_set()

        # All thing to do, requirement + extra if added via extra
        worklist: list[Tuple[inmanta.util.CanonicalRequirement, Optional[str]]] = []
        seen_requirements: set[inmanta.util.CanonicalRequirement] = set()

        reqs_as_requirements: Sequence[inmanta.util.CanonicalRequirement] = cls._get_as_requirements_type(requirements)
        for r in reqs_as_requirements:
            worklist.append((r, None))

        while worklist:
            r, contained_in_extra = worklist.pop()

            if r in seen_requirements:
                continue
            seen_requirements.add(r)

            # Requirements created by the `Distribution.requires()` method have the extra, the Requirement was created
            # from,
            # set as a marker. The line below makes sure that the "extra" marker matches. The marker is not set by
            # `Distribution.requires()` when the package is installed in editable mode, but setting it always doesn't make
            # the marker evaluation fail.
            environment_marker_evaluation = {"extra": contained_in_extra} if contained_in_extra else None
            if r.marker and not r.marker.evaluate(environment=environment_marker_evaluation):
                # The marker of the requirement doesn't apply on this environment
                continue

            name = NormalizedName(r.name)  # requirement is normalized
            if name not in installed_packages or not r.specifier.contains(installed_packages[name], prereleases=True):
                return False

            if r.extras:
                # if we have extr'as these may not have been installed!
                # We have to recurse for them!
                for extra in r.extras:
                    found_distribution: Optional[Distribution] = distribution(name)
                    if found_distribution is None:
                        return False

                    environment_marker_evaluation = {"extra": extra}
                    all_requires: list[inmanta.util.CanonicalRequirement] = [
                        inmanta.util.parse_requirement(requirement) for requirement in (found_distribution.requires or [])
                    ]
                    pkgs_required_by_extra: list[inmanta.util.CanonicalRequirement] = [
                        requirement
                        for requirement in all_requires
                        if requirement.marker and requirement.marker.evaluate(environment_marker_evaluation)
                    ]
                    for req in pkgs_required_by_extra:
                        worklist.append((req, extra))
        return True

    @classmethod
    def get_packages_in_working_set(cls, inmanta_modules_only: bool = False) -> dict[NormalizedName, packaging.version.Version]:
        """
        Return all package names (under the canonicalized form) present on the sys.path with their version

        :param inmanta_modules_only: Only return inmanta modules from the working set
        """
        return {
            name: packaging.version.Version(dist_info.version)
            for name, dist_info in cls.get_dist_in_working_set().items()
            if not inmanta_modules_only or name.startswith(const.MODULE_PKG_NAME_PREFIX)
        }

    @classmethod
    def get_dist_in_working_set(cls) -> dict[NormalizedName, Distribution]:
        """
        Return all packages (with the canonicalized name) present on the sys.path
        """
        return {
            packaging.utils.canonicalize_name(dist_info.name): dist_info
            for dist_info in reversed(list(distributions()))  # make sure we get the first entry for every name
        }

    @classmethod
    def get_dependency_tree(cls, dists: abc.Iterable[NormalizedName]) -> abc.Set[NormalizedName]:
        """
        Returns the full set of all dependencies (both direct and transitive) for the given distributions. Includes the
        distributions themselves.
        If one of the distributions or its dependencies is not installed, it is still included in the set but its dependencies
        are not.

        extras are ignored

        :param dists: The keys for the distributions to get the dependency tree for.
        """
        # create dict for O(1) lookup
        installed_distributions: abc.Mapping[NormalizedName, Distribution] = PythonWorkingSet.get_dist_in_working_set()

        def _get_tree_recursive(
            dists: abc.Iterable[NormalizedName], acc: abc.Set[NormalizedName] = frozenset()
        ) -> abc.Set[NormalizedName]:
            """
            :param acc: Accumulator for requirements that have already been recursed on.
            """
            return reduce(_get_tree_recursive_single, dists, acc)

        def _get_tree_recursive_single(acc: abc.Set[NormalizedName], dist: NormalizedName) -> abc.Set[NormalizedName]:
            if dist in acc:
                return acc

            if dist not in installed_distributions:
                return acc | {dist}

            # recurse on direct dependencies
            return _get_tree_recursive(
                (
                    requirement.name
                    for requirement in (
                        parse_requirement(raw_requirement) for raw_requirement in (installed_distributions[dist].requires or [])
                    )
                    if (requirement.marker is None or requirement.marker.evaluate())
                ),
                acc=acc | {dist},
            )

        return _get_tree_recursive(dists)


@dataclass
class LocalPackagePath:
    path: str
    editable: bool = False


class PipListFormat(enum.Enum):
    """
    The different output formats that can be passed to the `pip list` command.
    """

    columns = "columns"
    freeze = "freeze"
    json = "json"


class PipUpgradeStrategy(enum.Enum):
    """
    The upgrade strategy used by pip (`--upgrade-strategy` option). Determines upgrade behavior for dependencies of packages to
    upgrade.
    """

    EAGER = "eager"
    ONLY_IF_NEEDED = "only-if-needed"


def assert_pip_has_source(pip_config: PipConfig, reason: str) -> None:
    """Ensure this index has a valid package source, otherwise raise exception"""
    # placed here and not in pip_config to avoid import loop
    if not pip_config.has_source():
        raise PackageNotFound(
            f"Attempting to install {reason} but pip is not configured. Add the relevant pip "
            f"indexes to the project config file. e.g. to set PyPi as pip index, add the following "
            "to `project.yml`:"
            "\npip:"
            "\n  index_url: https://pypi.org/simple"
            "\nAnother option is to set `pip.use_system_config = true` to use the system's pip config."
        )


class PipCommandBuilder:
    """
    Class used to compose pip commands.
    """

    @classmethod
    def compose_uninstall_command(cls, python_path: str, pkg_names: Sequence[str]) -> list[str]:
        """
        Return the pip command to uninstall the given python packages.

        :param python_path: The python interpreter to use in the command.
        :param pkg_names: The names of the python packages that should be uninstalled.
        """
        return [python_path, "-m", "pip", "uninstall", "-y", *pkg_names]

    @classmethod
    def compose_list_command(
        cls, python_path: str, format: Optional[PipListFormat] = None, only_editable: bool = False
    ) -> list[str]:
        """
        Generate a `pip list` command for the given arguments.

        :param python_path: The python interpreter to use in the command.
        :param format: The output format to use.
        :param only_editable: Whether the output should only contain project installed in editable mode.
        """
        return [
            python_path,
            "-m",
            "pip",
            "list",
            # we disable pip-version check to prevent the json format from getting other output
            # deeply confusing issue: https://github.com/pypa/pip/issues/10715
            *(["--disable-pip-version-check", "--format", format.value] if format else []),
            *(["--editable"] if only_editable else []),
        ]


class Pip(PipCommandBuilder):
    """Class to handle interactions with pip"""

    @classmethod
    def run_pip_install_command_from_config(
        cls,
        python_path: str,
        config: PipConfig,
        requirements: Optional[Sequence[packaging.requirements.Requirement]] = None,
        requirements_files: Optional[list[str]] = None,
        upgrade: bool = False,
        upgrade_strategy: PipUpgradeStrategy = PipUpgradeStrategy.ONLY_IF_NEEDED,
        constraints_files: Optional[list[str]] = None,
        paths: Optional[list[LocalPackagePath]] = None,
    ) -> None:
        """
        Perform a pip install according to the given config

        :param python_path: the python path to use
        :param config: the pip config to use

        :param requirements: which requirements to install
        :param requirements_files: which requirements_files to install (-r)
        :param paths: which paths to install

        :param constraints_files: pass along the following constraint files

        :param upgrade: make pip do an upgrade
        :param upgrade_strategy: what upgrade strategy to use
        """

        cmd, constraints_files_clean, requirements_files_clean, sub_env = cls._prepare_pip_install_command(
            python_path,
            config,
            requirements,
            requirements_files,
            upgrade,
            upgrade_strategy,
            constraints_files,
            paths,
        )

        cls.run_pip(cmd, sub_env, constraints_files_clean, requirements_files_clean)

    @classmethod
    async def async_run_pip_install_command_from_config(
        cls,
        python_path: str,
        config: PipConfig,
        requirements: Optional[Sequence[packaging.requirements.Requirement]] = None,
        requirements_files: Optional[list[str]] = None,
        upgrade: bool = False,
        upgrade_strategy: PipUpgradeStrategy = PipUpgradeStrategy.ONLY_IF_NEEDED,
        constraints_files: Optional[list[str]] = None,
        paths: Optional[list[LocalPackagePath]] = None,
    ) -> None:
        """
        Perform a pip install according to the given config

        :param python_path: the python path to use
        :param config: the pip config to use

        :param requirements: which requirements to install
        :param requirements_files: which requirements_files to install (-r)
        :param paths: which paths to install

        :param constraints_files: pass along the following constraint files

        :param upgrade: make pip do an upgrade
        :param upgrade_strategy: what upgrade strategy to use
        """

        cmd, constraints_files_clean, requirements_files_clean, sub_env = cls._prepare_pip_install_command(
            python_path,
            config,
            requirements,
            requirements_files,
            upgrade,
            upgrade_strategy,
            constraints_files,
            paths,
        )
        await cls.async_run_pip(cmd, sub_env, constraints_files_clean, requirements_files_clean)

    @classmethod
    def run_pip_download_command(
        cls,
        python_path: str,
        dependencies: Sequence[inmanta.util.CanonicalRequirement],
        output_dir: str,
        constraints: Sequence[inmanta.util.CanonicalRequirement] | None = None,
        pip_config: PipConfig | None = None,
        no_deps: bool = False,
        no_binary: str | None = None,
    ) -> None:
        if not dependencies:
            return

        index_args: list[str]
        env_vars: dict[str, str]
        if pip_config:
            index_args = pip_config.get_index_args()
            env_vars = pip_config.get_environment_variables()
        else:
            index_args = []
            env_vars = os.environ.copy()

        cmd = [python_path, "-m", "pip", "download"]
        if index_args:
            cmd.extend(index_args)
        if no_deps:
            cmd.append("--no-deps")
        if no_binary:
            cmd.extend(["--no-binary", no_binary])
        with tempfile.NamedTemporaryFile() as fd:
            if constraints:
                fd.write("\n".join(str(c) for c in constraints).encode())
                fd.seek(0)
                cmd.extend(["-c", fd.name])
            if dependencies:
                cmd.extend([str(r) for r in dependencies])
            cls.run_pip(cmd, env=env_vars, cwd=output_dir, constraints_files=[fd.name])

    @classmethod
    def _prepare_pip_install_command(
        cls,
        python_path: str,
        config: PipConfig,
        requirements: Optional[Sequence[packaging.requirements.Requirement]] = None,
        requirements_files: Optional[list[str]] = None,
        upgrade: bool = False,
        upgrade_strategy: PipUpgradeStrategy = PipUpgradeStrategy.ONLY_IF_NEEDED,
        constraints_files: Optional[list[str]] = None,
        paths: Optional[list[LocalPackagePath]] = None,
    ) -> Tuple[list[str], list[str], list[str], dict[str, str]]:
        # What
        requirements = requirements if requirements is not None else []
        clean_requirements_files = requirements_files if requirements_files is not None else []
        paths = paths if paths is not None else []
        local_paths: Iterator[LocalPackagePath] = (
            # make sure we only try to install from a local source: add leading `./` and trailing `/` to explicitly tell pip
            # we're pointing to a local directory.
            LocalPackagePath(path=os.path.join(".", path.path, ""), editable=path.editable)
            for path in paths
        )
        install_args = [
            *(str(requirement) for requirement in requirements),
            *chain.from_iterable(["-r", f] for f in clean_requirements_files),
            *chain.from_iterable(["-e", path.path] if path.editable else [path.path] for path in local_paths),
        ]
        # From where
        if paths:
            # For local installs, we allow not having an index set.
            pass
        else:
            # All others need an index
            assert_pip_has_source(config, "'" + " ".join(install_args) + "'")
        index_args: list[str] = config.get_index_args()
        clean_constraints_files = constraints_files if constraints_files is not None else []
        # Command
        cmd = [
            python_path,
            "-m",
            "pip",
            "install",
            *(["--upgrade", "--upgrade-strategy", upgrade_strategy.value] if upgrade else []),
            *(["--pre"] if config.pre else []),
            *chain.from_iterable(["-c", f] for f in clean_constraints_files),
            *install_args,
            *index_args,
        ]
        # ISOLATION!
        sub_env = config.get_environment_variables()
        return cmd, clean_constraints_files, clean_requirements_files, sub_env

    @classmethod
    def run_pip(
        cls,
        cmd: list[str],
        env: dict[str, str],
        constraints_files: list[str] | None = None,
        requirements_files: list[str] | None = None,
        cwd: str | None = None,
    ) -> None:
        if constraints_files is None:
            constraints_files = []
        if requirements_files is None:
            requirements_files = []
        cls._log_before_run(cmd, constraints_files, requirements_files)
        return_code, full_output = CommandRunner(LOGGER_PIP).run_command_and_stream_output(cmd, env_vars=env, cwd=cwd)
        cls._process_return(cmd, env, full_output, return_code)

    @classmethod
    async def async_run_pip(
        cls, cmd: list[str], env: dict[str, str], constraints_files: list[str], requirements_files: list[str]
    ) -> None:
        cls._log_before_run(cmd, constraints_files, requirements_files)
        return_code, full_output = await CommandRunner(LOGGER_PIP).async_run_command_and_stream_output(cmd, env_vars=env)
        cls._process_return(cmd, env, full_output, return_code)

    @classmethod
    def _log_before_run(cls, cmd: list[str], constraints_files: list[str], requirements_files: list[str]) -> None:
        def create_log_content_files(title: str, files: list[str]) -> list[str]:
            """
            Log the content of a list of files with indentations in the following format:

            Content of [title]:
                [files[0]]:
                    line 1 in files[0]
                [files[1]]:
                    line 1 in files[1]
                    line 2 in files[1]
                    line 3 in files[1]
                    ...
                [files[2]]:
                ...

            this function will skip empty lines in files
            """
            log_msg: list[str] = [f"Content of {title}:\n"]
            indentation: str = "    "
            for file in files:
                log_msg.append(indent(file + ":\n", indentation))
                with open(file) as f:
                    for line in f:
                        if line.strip():
                            log_msg.append(indent(line.strip() + "\n", 2 * indentation))
            return log_msg

        log_msg: list[str] = []
        if requirements_files:
            log_msg.extend(create_log_content_files("requirements files", requirements_files))
        if constraints_files:
            log_msg.extend(create_log_content_files("constraints files", constraints_files))
        log_msg.append("Pip command: " + " ".join(cmd))
        LOGGER_PIP.debug("".join(log_msg).strip())

    @classmethod
    def _process_return(cls, cmd: list[str], env: dict[str, str], full_output: list[str], return_code: int) -> None:
        if return_code != 0:
            not_found: list[str] = []
            conflicts: list[str] = []
            indexes: str = ""
            for line in full_output:
                m = re.search(r"No matching distribution found for ([\S]+)", line)
                if m:
                    # Add missing package name to not_found list
                    not_found.append(m.group(1))

                if "versions have conflicting dependencies" in line:
                    conflicts.append(line)
                # Get the indexes line from full_output
                # This is not printed when not using any index or when only using PyPi
                if "Looking in indexes:" in line:
                    indexes = line
            if not_found:
                no_index: bool = "--no-index" in cmd or strtobool(env.get("PIP_NO_INDEX", "false"))
                if no_index:
                    msg = "Packages %s were not found. No indexes were used." % ", ".join(not_found)
                elif indexes:
                    msg = "Packages %s were not found in the given indexes. (%s)" % (", ".join(not_found), indexes)
                else:
                    msg = "Packages %s were not found at PyPI." % ", ".join(not_found)
                raise PackageNotFound(msg)
            if conflicts:
                raise ConflictingRequirements("\n".join(conflicts))
            raise PipInstallError(
                f"Process {cmd} exited with return code {return_code}. "
                "Increase the verbosity level with the -v option for more information."
            )


class PythonEnvironment:
    """
    A generic Python environment.

    The implementation of this class is based on the invariant that the version of the inmanta-core and the
    Inmanta product packages don't change.
    """

    _invalid_chars_in_path_re = re.compile(r'["$`]')

    def __init__(self, *, env_path: Optional[str] = None, python_path: Optional[str] = None) -> None:
        if (env_path is None) == (python_path is None):
            raise ValueError("Exactly one of `env_path` and `python_path` needs to be specified")
        self.env_path: str
        self.python_path: str
        self._parent_python: Optional[str] = None
        if env_path is not None:
            self.env_path = env_path
            self.python_path = self.get_python_path_for_env_path(self.env_path)
            if not self.env_path:
                raise ValueError("The env_path cannot be an empty string.")
        else:
            assert python_path is not None
            self.python_path = python_path
            self.env_path = self.get_env_path_for_python_path(self.python_path)
            if not self.python_path:
                raise ValueError("The python_path cannot be an empty string.")
        self.validate_path(self.env_path)
        self.site_packages_dir: str = self.get_site_dir_for_env_path(self.env_path)
        self._path_pth_file = os.path.join(self.site_packages_dir, "inmanta-inherit-from-parent-venv.pth")

    def validate_path(self, path: str) -> None:
        """
        The given path is used in the `./bin/activate` file of the created venv without escaping any special characters.
        As such, we refuse all special characters here that might cause the given path to be interpreted incorrectly:

            * $: Character used for variable expansion in bash strings.
            * `: Character used to perform command substitution in bash strings.
            * ": Character that will be interpreted incorrectly as the end of the string.

        :param path: Path to validate.
        """
        if not path:
            raise ValueError("Cannot create virtual environment because the provided path is an empty string.")

        match = PythonEnvironment._invalid_chars_in_path_re.search(path)
        if match:
            raise ValueError(
                f"Cannot create virtual environment because the provided path `{path}` contains an"
                f" invalid character (`{match.group()}`)."
            )

    @classmethod
    def get_python_path_for_env_path(cls, env_path: str) -> str:
        """
        For the given venv directory (`env_path`) return the path to the Python interpreter.
        """
        python_name: str = os.path.basename(sys.executable)
        return (
            os.path.join(env_path, "Scripts", python_name)
            if sys.platform == "win32"
            else os.path.join(env_path, "bin", python_name)
        )

    @classmethod
    def get_site_dir_for_env_path(cls, env_path: str) -> str:
        """
        Return the site directory for a given venv directory.
        """
        return (
            os.path.join(env_path, "Lib", "site-packages")
            if sys.platform == "win32"
            else os.path.join(
                env_path, "lib", "python%s" % ".".join(str(digit) for digit in sys.version_info[:2]), "site-packages"
            )
        )

    @classmethod
    def get_env_path_for_python_path(cls, python_path: str) -> str:
        """
        For a given path to a python binary, return the path to the venv directory.
        """
        return os.path.dirname(os.path.dirname(python_path))

    def init_env(self) -> None:
        """
        Initialize the virtual environment.
        """
        self._parent_python = sys.executable
        LOGGER.info("Initializing virtual environment at %s", self.env_path)

        # check if the virtual env exists
        if os.path.isdir(self.env_path) and os.listdir(self.env_path):
            self.can_activate()
        else:
            path = os.path.realpath(self.env_path)
            try:
                venv.create(path, clear=True, with_pip=False)
                self._write_pip_binary()
                self._write_pth_file()
            except CalledProcessError as e:
                raise VenvCreationFailedError(msg=f"Unable to create new virtualenv at {self.env_path} ({e.stdout.decode()})")
            except Exception:
                raise VenvCreationFailedError(msg=f"Unable to create new virtualenv at {self.env_path}")
            LOGGER.debug("Created a new virtualenv at %s", self.env_path)

        if not os.path.exists(self._path_pth_file):
            # Venv was created using an older version of Inmanta -> Update pip binary and set sitecustomize.py file
            self._write_pip_binary()
            self._write_pth_file()

    def can_activate(self) -> None:
        """
        Can this venv be activated with this current python version?

        raises a VenvActivationFailedError exception if this is not the case
        """
        # Make sure the venv hosts the same python version as the running process
        if sys.platform.startswith("linux"):
            # Check if the python binary exists in the environment's bin directory
            if not os.path.exists(self.python_path):
                raise VenvActivationFailedError(
                    msg=f"Unable to use virtualenv at {self.env_path} as no Python installation exists."
                )
            # On linux based systems, the python version is in the path to the site packages dir:
            if not os.path.exists(self.site_packages_dir):
                raise VenvActivationFailedError(
                    msg=f"Unable to use virtualenv at {self.env_path} because its Python version "
                    "is different from the Python version of this process."
                )
        else:
            # On other distributions a more costly check is required:
            # Get version as a (major, minor) tuple for the venv and the running process
            venv_python_version = subprocess.check_output([self.python_path, "--version"]).decode("utf-8").strip().split()[1]
            venv_python_version = tuple(map(int, venv_python_version.split(".")))[:2]

            running_process_python_version = sys.version_info[:2]

            if venv_python_version != running_process_python_version:
                raise VenvActivationFailedError(
                    msg=f"Unable to use virtualenv at {self.env_path} because its Python version "
                    "is different from the Python version of this process."
                )

    def _write_pip_binary(self) -> None:
        """
        write out a "stub" pip binary so that pip list works in the virtual env.
        """
        pip_path = os.path.join(self.env_path, "bin", "pip")

        with open(pip_path, "w", encoding="utf-8") as fd:
            fd.write(
                """#!/bin/sh
"$(dirname "$0")/python" -m pip $@
                """.strip()
            )
        os.chmod(pip_path, 0o755)

    def _write_pth_file(self) -> None:
        """
        Write an inmanta-inherit-from-parent-venv.pth file to the venv to ensure that an activation of this venv will also
        activate the parent venv. The site directories of the parent venv should appear later in sys.path than the ones of
        this venv.
        """
        site_dir_strings: list[str] = ['"' + p.replace('"', r"\"") + '"' for p in list(sys.path)]
        add_site_dir_statements: str = "\n".join(
            [f"site.addsitedir({p}) if {p} not in sys.path else None" for p in site_dir_strings]
        )
        script = f"""
import os
import site
import sys


# Ensure inheritance from all parent venvs + process their .pth files
{add_site_dir_statements}
        """
        script_as_oneliner = "; ".join(
            [line for line in script.split("\n") if line.strip() and not line.strip().startswith("#")]
        )
        with open(self._path_pth_file, "w", encoding="utf-8") as fd:
            fd.write(script_as_oneliner)

    def get_installed_packages(self, only_editable: bool = False) -> dict[NormalizedName, packaging.version.Version]:
        """
        Return a list of all installed packages in the site-packages of a python interpreter.

        :param only_editable: List only packages installed in editable mode.
        :return: A dict with package names as keys and versions as values
        """
        cmd = PipCommandBuilder.compose_list_command(self.python_path, format=PipListFormat.json, only_editable=only_editable)
        output = CommandRunner(LOGGER_PIP).run_command_and_log_output(cmd, stderr=subprocess.DEVNULL, env=os.environ.copy())
        return {canonicalize_name(r["name"]): packaging.version.Version(r["version"]) for r in json.loads(output)}

    def download_distributions(
        self,
        output_directory: str,
        pip_config: PipConfig | None,
        dependencies: Sequence[inmanta.util.CanonicalRequirement],
        constraints: Sequence[inmanta.util.CanonicalRequirement] | None = None,
        no_binary: str | None = None,
        no_deps: bool = False,
    ) -> None:
        """
        Download the python distribution packages that satisfy the given requirements.
        """
        if not dependencies:
            return
        Pip.run_pip_download_command(
            python_path=self.python_path,
            output_dir=output_directory,
            pip_config=pip_config,
            dependencies=dependencies,
            constraints=constraints,
            no_deps=no_deps,
            no_binary=no_binary,
        )

    def install_for_config(
        self,
        requirements: Sequence[inmanta.util.CanonicalRequirement],
        config: PipConfig,
        upgrade: bool = False,
        constraint_files: Optional[list[str]] = None,
        upgrade_strategy: PipUpgradeStrategy = PipUpgradeStrategy.ONLY_IF_NEEDED,
        paths: list[LocalPackagePath] = [],
        add_inmanta_requires: bool = True,
        constraints: Sequence[inmanta.util.CanonicalRequirement] | None = None,
    ) -> None:
        """
        Perform a pip install in this environment, according to the given config

        :param requirements: which requirements to install
        :param paths: which paths to install
        :param config: the pip config to use
        :param constraint_files: pass along the following constraint files
        :param upgrade: make pip do an upgrade
        :param upgrade_strategy: what upgrade strategy to use

        limitation:
         - When upgrade is false, if requirements are already installed constraints from constraint files may not be verified.
        """
        if len(requirements) == 0 and len(paths) == 0:
            raise Exception("install_for_config requires at least one requirement or path to install")
        constraint_files = constraint_files if constraint_files is not None else []
        if add_inmanta_requires:
            inmanta_requirements = self._get_requirements_on_inmanta_package()
        else:
            inmanta_requirements = []

        with tempfile.NamedTemporaryFile() as fd:
            if constraints:
                fd.write("\n".join(str(c) for c in constraints).encode())
                fd.seek(0)
            Pip.run_pip_install_command_from_config(
                python_path=self.python_path,
                config=config,
                requirements=[*requirements, *inmanta_requirements],
                constraints_files=[*constraint_files, fd.name],
                upgrade=upgrade,
                upgrade_strategy=upgrade_strategy,
                paths=paths,
            )

    async def async_install_for_config(
        self,
        requirements: list[packaging.requirements.Requirement],
        config: PipConfig,
        upgrade: bool = False,
        constraint_files: Optional[list[str]] = None,
        upgrade_strategy: PipUpgradeStrategy = PipUpgradeStrategy.ONLY_IF_NEEDED,
        paths: list[LocalPackagePath] = [],
    ) -> None:
        """
        Perform a pip install in this environment, according to the given config

        :param requirements: which requirements to install
        :param config: the pip config to use
        :param upgrade: make pip do an upgrade
        :param constraint_files: pass along the following constraint files
        :param upgrade_strategy: what upgrade strategy to use
        :param paths: which paths to install

        limitation:
         - When upgrade is false, if requirements are already installed constraints from constraint files may not be verified.
        """
        if len(requirements) == 0 and len(paths) == 0:
            raise Exception("install_for_config requires at least one requirement or path to install")
        constraint_files = constraint_files if constraint_files is not None else []
        inmanta_requirements = self._get_requirements_on_inmanta_package()

        await Pip.async_run_pip_install_command_from_config(
            python_path=self.python_path,
            config=config,
            requirements=[*requirements, *inmanta_requirements],
            constraints_files=constraint_files,
            upgrade=upgrade,
            upgrade_strategy=upgrade_strategy,
            paths=paths,
        )

    def install_from_index(
        self,
        requirements: list[inmanta.util.CanonicalRequirement],
        index_urls: Optional[list[str]] = None,
        upgrade: bool = False,
        allow_pre_releases: bool = False,
        constraint_files: Optional[list[str]] = None,
        upgrade_strategy: PipUpgradeStrategy = PipUpgradeStrategy.ONLY_IF_NEEDED,
        use_pip_config: Optional[bool] = False,
    ) -> None:
        """This method provides backward compatibility with ISO6"""
        if len(requirements) == 0:
            raise Exception("install_from_index requires at least one requirement to install")

        if not index_urls:
            index_url = None
            extra_index_url = []
        else:
            index_url = index_urls[0]
            extra_index_url = index_urls[1:]

        self.install_for_config(
            requirements=requirements,
            config=PipConfig(
                index_url=index_url,
                extra_index_url=extra_index_url,
                pre=allow_pre_releases,
                use_system_config=use_pip_config if use_pip_config is not None else True,
            ),
            upgrade=upgrade,
            constraint_files=constraint_files,
            upgrade_strategy=upgrade_strategy,
        )

    def install_from_source(
        self,
        paths: list[LocalPackagePath],
        constraint_files: Optional[list[str]] = None,
    ) -> None:
        """
        Install one or more packages from source. Any path arguments should be local paths to a package directory or wheel.

        This method provides backward compatibility with ISO6
        """
        if len(paths) == 0:
            raise Exception("install_from_source requires at least one package to install")
        self.install_for_config(
            requirements=[],
            paths=paths,
            config=LEGACY_PIP_DEFAULT,
            constraint_files=constraint_files,
        )

    def install_from_list(
        self,
        requirements_list: Sequence[str],
        *,
        upgrade: bool = False,
        upgrade_strategy: PipUpgradeStrategy = PipUpgradeStrategy.ONLY_IF_NEEDED,
        use_pip_config: Optional[bool] = False,
    ) -> None:
        """
        Install requirements from a list of requirement strings. This method uses the Python package repositories
        configured on the host.
        :param requirements_list: List of requirement strings to install.
        :param upgrade: Upgrade requirements to the latest compatible version.
        :param upgrade_strategy: The upgrade strategy to use for requirements' dependencies.
        :param use_pip_config: ignored

        This method provides backward compatibility with ISO6
        use_pip_config was ignored on ISO6 and it still is
        """
        self.install_from_index(
            requirements=inmanta.util.parse_requirements(requirements_list),
            upgrade=upgrade,
            upgrade_strategy=upgrade_strategy,
            use_pip_config=True,
        )

    @classmethod
    def get_protected_inmanta_packages(cls) -> list[str]:
        """
        Returns the list of packages that should not be installed/updated by any operation on a Python environment.
        This list of packages will be under the canonical form.
        """
        return [
            # Protect product packages
            packaging.utils.canonicalize_name("inmanta"),
            packaging.utils.canonicalize_name("inmanta-service-orchestrator"),
            # Protect all server extensions
            *(
                packaging.utils.canonicalize_name(f"inmanta-{ext_name}")
                for ext_name in InmantaBootloader.get_available_extensions().keys()
            ),
        ]

    @classmethod
    def _get_requirements_on_inmanta_package(cls) -> Sequence[inmanta.util.CanonicalRequirement]:
        """
        Returns the content of the requirement file that should be supplied to each `pip install` invocation
        to make sure that no Inmanta packages gets overridden.
        """
        protected_inmanta_packages: list[str] = cls.get_protected_inmanta_packages()
        workingset: dict[NormalizedName, packaging.version.Version] = PythonWorkingSet.get_packages_in_working_set()
        return [
            inmanta.util.parse_requirement(requirement=f"{pkg}=={workingset[pkg]}")
            for pkg in workingset
            if pkg in protected_inmanta_packages
        ]


class CommandRunner:
    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger

    def run_command_and_log_output(
        self, cmd: list[str], env: Optional[dict[str, str]] = None, stderr: Optional[int] = None, cwd: Optional[str] = None
    ) -> str:
        output: bytes = b""  # Make sure the var is always defined in the except bodies
        try:
            output = subprocess.check_output(cmd, stderr=stderr, env=env, cwd=cwd)
        except CalledProcessError as e:
            if e.stderr:
                msg = e.stderr.decode()
            elif e.output:
                msg = e.output.decode()
            else:
                msg = ""
            self.logger.error("%s: %s", cmd, msg)
            raise
        except Exception:
            self.logger.error("%s: %s", cmd, output.decode())
            raise
        else:
            self.logger.debug("%s: %s", cmd, output.decode())
            return output.decode()

    def run_command_and_stream_output(
        self,
        cmd: list[str],
        timeout: float = 10,
        env_vars: Optional[Mapping[str, str]] = None,
        cwd: str | None = None,
    ) -> tuple[int, list[str]]:
        """
        Similar to the _run_command_and_log_output method, but here, the output is logged on the fly instead of at the end
        of the sub-process.
        """
        full_output: list[str] = []
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env_vars,
            cwd=cwd,
        )
        assert process.stdout is not None  # Make mypy happy
        try:
            for line in process.stdout:
                # Eagerly consume the buffer to avoid a deadlock in case the subprocess fills it entirely.
                output = line.decode().strip()
                full_output.append(output)
                self.logger.debug(output)
        finally:
            process.stdout.close()

        try:
            return_code = process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            process.kill()
            return -1, full_output
        else:
            return return_code, full_output

    async def async_run_command_and_stream_output(
        self,
        cmd: list[str],
        timeout: float = 10,
        env_vars: Optional[Mapping[str, str]] = None,
    ) -> tuple[int, list[str]]:
        """
        Similar to the _run_command_and_log_output method, but here, the output is logged on the fly instead of at the end
        of the sub-process.
        """
        full_output = []
        # We use close_fds here to avoid
        # the bug https://github.com/python/cpython/issues/103911#issuecomment-2333963137
        # We attempt to get on the code path that uses _posix_spawn instead of _fork_exec
        # Subprocess shell did not work to prevent it
        # Improved escaping (shlex.quote) to prevent `>` from leaking did not help
        process = await asyncio.create_subprocess_exec(
            *cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env_vars, close_fds=False
        )
        assert process.stdout is not None  # Make mypy happy
        async for line in process.stdout:
            # Eagerly consume the buffer to avoid a deadlock in case the subprocess fills it entirely.
            output = line.decode().strip()
            full_output.append(output)
            self.logger.debug(output)

        try:
            return_code = await asyncio.wait_for(process.wait(), timeout=timeout)
        except TimeoutError:
            process.kill()
            return -1, full_output
        else:
            return return_code, full_output


class ActiveEnv(PythonEnvironment):
    """
    The active Python environment. Method implementations assume this environment is active when they're called.
    Activating another environment that inherits from this one is allowed.
    """

    _egg_fragment_re = re.compile(r"#egg=(?P<name>[^&]*)")
    _at_fragment_re = re.compile(r"^(?P<name>[^@]+)@(?P<req>.+)")

    def __init__(self, *, env_path: Optional[str] = None, python_path: Optional[str] = None) -> None:
        super().__init__(env_path=env_path, python_path=python_path)

    def is_using_virtual_env(self) -> bool:
        return True

    def use_virtual_env(self) -> None:
        """
        Activate the virtual environment.
        """
        return

    def are_installed(self, requirements: req_list) -> bool:
        """
        Return True iff the given requirements are installed in this environment.
        """
        assert self.is_using_virtual_env()
        return PythonWorkingSet.are_installed(requirements)

    def install_for_config(
        self,
        requirements: Sequence[inmanta.util.CanonicalRequirement],
        config: PipConfig,
        upgrade: bool = False,
        constraint_files: Optional[list[str]] = None,
        upgrade_strategy: PipUpgradeStrategy = PipUpgradeStrategy.ONLY_IF_NEEDED,
        paths: list[LocalPackagePath] = [],
        add_inmanta_requires: bool = True,
        constraints: Sequence[inmanta.util.CanonicalRequirement] | None = None,
    ) -> None:
        if (not upgrade and self.are_installed(requirements)) and not paths:
            return
        try:
            super().install_for_config(
                requirements,
                config,
                upgrade,
                constraint_files,
                upgrade_strategy,
                paths,
                add_inmanta_requires,
                constraints=constraints,
            )
        finally:
            self.notify_change()

    @classmethod
    def get_module_file(cls, module: str) -> Optional[tuple[Optional[str], Loader]]:
        """
        Get the location of the init file for a Python module within the active environment. Returns the file path as observed
        by Python. For editable installs, this may or may not be a symlink to the actual location (see implementation
        mechanisms in setuptools docs: https://setuptools.pypa.io/en/latest/userguide/development_mode.html).

        :return: A tuple of the path and the associated loader, if the module is found.
        """
        spec: Optional[ModuleSpec]
        try:
            spec = importlib.util.find_spec(module)
        # inmanta.loader.PluginModuleLoader raises ImportError if module is not found
        except (ImportError, ModuleNotFoundError):
            spec = None
        return (spec.origin, spec.loader) if spec is not None else None

    def get_installed_packages(self, only_editable: bool = False) -> dict[NormalizedName, packaging.version.Version]:
        """
        Return a list of all installed packages in the site-packages of a python interpreter.

        :param only_editable: List only packages installed in editable mode.
        :return: A dict with package names as keys and versions as values
        """
        if self.is_using_virtual_env() and not only_editable:
            return PythonWorkingSet.get_packages_in_working_set()
        return super().get_installed_packages(only_editable=only_editable)

    def notify_change(self) -> None:
        """
        This method must be called when a package is installed or removed from the environment in order for Python to detect
        the change. Namespace packages installed in editable mode in particular require this method to allow them to be found by
        get_module_file().
        """
        # Make sure that the .pth files in the site-packages directory are processed.
        # This is required to make editable installs work.
        site.addsitedir(self.site_packages_dir)
        importlib.invalidate_caches()

        if const.PLUGINS_PACKAGE in sys.modules:
            mod = sys.modules[const.PLUGINS_PACKAGE]
            if mod is not None:
                # Make mypy happy
                assert mod.__spec__.submodule_search_locations is not None
                if self.site_packages_dir not in mod.__spec__.submodule_search_locations and os.path.exists(
                    os.path.join(self.site_packages_dir, const.PLUGINS_PACKAGE)
                ):
                    """
                    A V2 module was installed in this virtual environment, but the inmanta_plugins package was already
                    loaded before this venv was activated. Reload the inmanta_plugins package to ensure that all V2 modules
                    installed in this virtual environment are discovered correctly.

                    This is required to cover the following scenario:

                        * Two venvs are stacked on top of each other. The parent venv contains the inmanta-core package and
                          the subvenv is empty.
                        * The inmanta_plugins package gets loaded before a V2 module is installed in the subvenv. This way,
                          the module object in sys.modules, doesn't have the site dir of the subvenv in its
                          submodule_search_locations. This field caches where the loader should look for the namespace packages
                          that are part of the inmanta_plugins namespace.
                        * When a V2 module gets installed in the subvenv now, the loader will not find the newly installed V2
                          module, because it's not considering the site dir of the subvenv.

                    The above-mentioned scenario can only be triggered by test cases, because:
                        1) The compiler venv was removed. As such, no new venv are activated on the fly by production code
                           paths.
                        2) The compiler service creates a new subvenv for each inmanta environment, but the inmanta commands
                           are executed in a subprocess.
                    """
                    importlib.reload(mod)


process_env: ActiveEnv = ActiveEnv(python_path=sys.executable)
"""
Singleton representing the Python environment this process is running in.

Should not be imported directly, as it can be updated
"""


@stable_api
def mock_process_env(*, python_path: Optional[str] = None, env_path: Optional[str] = None) -> None:
    """
    Overrides the process environment information. This forcefully sets the environment that is recognized as the outer Python
    environment. This function should only be called when a Python environment has been set up dynamically and this environment
    should be treated as if this process was spawned from it, and even then with great care.
    :param python_path: The path to the python binary. Only one of `python_path` and `env_path` should be set.
    :param env_path: The path to the python environment directory. Only one of `python_path` and `env_path` should be set.

    When using this method in a fixture to set and reset virtualenv, it is preferable to use `store_venv()`
    """
    process_env.__init__(python_path=python_path, env_path=env_path)  # type: ignore


def swap_process_env(env: ActiveEnv) -> ActiveEnv:
    """
    Overrides the process environment information.

    Returns the old active env

    For use in testing. Test as expected to swap the old process env back in place
    """
    global process_env
    old_env = process_env
    process_env = env
    return old_env


@stable_api
class VirtualEnv(ActiveEnv):
    """
    Creates and uses a virtual environment for this process. This virtualenv inherits from the previously active one.
    """

    def __init__(self, env_path: str) -> None:
        super().__init__(env_path=env_path)
        self.env_path: str = env_path
        self.virtual_python: Optional[str] = None
        self._using_venv: bool = False

    def exists(self) -> bool:
        """
        Returns True iff the venv exists on disk.
        """
        return os.path.exists(self.python_path)

    def is_using_virtual_env(self) -> bool:
        return self._using_venv

    def use_virtual_env(self) -> None:
        """
        Activate the virtual environment.
        """
        if self._using_venv:
            raise Exception(f"Already using venv {self.env_path}.")
        if not self.env_path:
            raise Exception("The env_path cannot be an empty string.")

        self.init_env()
        mock_process_env(python_path=self.python_path)

        self._activate_that()

        self._using_venv = True

    def _update_sys_path(self) -> None:
        """
        Updates sys.path by adding self.site_packages_dir. This method ensures
        that .pth files are processed.
        """
        prev_sys_path = list(sys.path)
        site.addsitedir(self.site_packages_dir)
        # Move the added items to the front of the path
        new_sys_path = [e for e in list(sys.path) if e not in prev_sys_path]
        new_sys_path += prev_sys_path
        # Set sys.path
        sys.path = new_sys_path

    def _activate_that(self) -> None:
        # adapted from https://github.com/pypa/virtualenv/blob/master/virtualenv_embedded/activate_this.py
        # MIT license
        # Copyright (c) 2007 Ian Bicking and Contributors
        # Copyright (c) 2009 Ian Bicking, The Open Planning Project
        # Copyright (c) 2011-2016 The virtualenv developers

        binpath: str = os.path.dirname(self.python_path)
        base: str = os.path.dirname(binpath)
        old_os_path = os.environ.get("PATH", "")
        os.environ["PATH"] = binpath + os.pathsep + old_os_path

        is_change = sys.prefix != base
        sys.real_prefix = sys.prefix
        sys.prefix = base
        self._update_sys_path()
        if is_change:
            self.notify_change()

    def install_for_config(
        self,
        requirements: Sequence[inmanta.util.CanonicalRequirement],
        config: PipConfig,
        upgrade: bool = False,
        constraint_files: Optional[list[str]] = None,
        upgrade_strategy: PipUpgradeStrategy = PipUpgradeStrategy.ONLY_IF_NEEDED,
        paths: list[LocalPackagePath] = [],
        add_inmanta_requires: bool = True,
        constraints: Sequence[inmanta.util.CanonicalRequirement] | None = None,
    ) -> None:
        if not self._using_venv:
            raise Exception(f"Not using venv {self.env_path}. use_virtual_env() should be called first.")
        super().install_for_config(
            requirements,
            config,
            upgrade,
            constraint_files,
            upgrade_strategy,
            paths,
            add_inmanta_requires,
            constraints=constraints,
        )


class VenvCreationFailedError(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)
        self.msg = msg


class VenvActivationFailedError(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)
        self.msg = msg


@dataclass
class VenvSnapshot:
    old_os_path: str
    old_prefix: str
    old_path: list[str]
    old_meta_path: "list[MetaPathFinderProtocol]"
    old_path_hooks: "list[Callable[[str], PathEntryFinderProtocol]]"
    old_pythonpath: Optional[str]
    old_os_venv: Optional[str]
    old_process_env_path: str
    old_process_env: ActiveEnv

    def restore(self) -> None:
        if sys.prefix != self.old_prefix:
            # only reset on folder change
            os.environ["PATH"] = self.old_os_path
            sys.prefix = self.old_prefix
            sys.path = self.old_path
            # reset sys.meta_path because it might contain finders for editable installs, make sure to keep the same object
            sys.meta_path.clear()
            sys.meta_path.extend(self.old_meta_path)
            sys.path_hooks.clear()
            sys.path_hooks.extend(self.old_path_hooks)
            # Clear cache for sys.path_hooks
            sys.path_importer_cache.clear()
            # Restore PYTHONPATH
            if self.old_pythonpath is not None:
                os.environ["PYTHONPATH"] = self.old_pythonpath
            elif "PYTHONPATH" in os.environ:
                del os.environ["PYTHONPATH"]
            # Restore VIRTUAL_ENV
            if self.old_os_venv is not None:
                os.environ["VIRTUAL_ENV"] = self.old_os_venv
            elif "VIRTUAL_ENV" in os.environ:
                del os.environ["VIRTUAL_ENV"]

        # We reset the process_env both ways: we put the reference back and we do an in_place update
        swap_process_env(self.old_process_env)
        mock_process_env(env_path=self.old_process_env_path)


def store_venv() -> VenvSnapshot:
    """
    Create a snapshot of the venv environment, for use in testing, to resest the test
    """

    self = VenvSnapshot(
        old_os_path=os.environ.get("PATH", ""),
        old_prefix=sys.prefix,
        old_path=list(sys.path),
        old_meta_path=sys.meta_path.copy(),
        old_path_hooks=sys.path_hooks.copy(),
        old_pythonpath=os.environ.get("PYTHONPATH", None),
        old_os_venv=os.environ.get("VIRTUAL_ENV", None),
        old_process_env=process_env,
        old_process_env_path=process_env.env_path,
    )
    return self



"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection

from inmanta import data


async def update(connection: Connection) -> None:
    """
    * Create index on table to efficiently filter resources based on the resource set that they belong to.
    * Ensure that the configurationmodel table has the is_suitable_for_partial_compiles column and make sure
      it's correctly populated for all existing models in the table.
    """
    # Add index on resource table
    await connection.execute("CREATE INDEX ON public.resource (environment, model, resource_set)")

    # Create column
    await connection.execute("ALTER TABLE public.configurationmodel ADD COLUMN is_suitable_for_partial_compiles boolean")
    # Populate column
    await connection.execute(
        f"""
            UPDATE {data.ConfigurationModel.table_name()}
            SET is_suitable_for_partial_compiles=FALSE
        """
    )
    # Add non null constraint
    await connection.execute("ALTER TABLE public.configurationmodel ALTER COLUMN is_suitable_for_partial_compiles SET NOT NULL")

    # Remove the version field from the attributes of a resource
    await connection.execute(
        f"""
            UPDATE {data.Resource.table_name()}
            SET attributes = attributes - 'version'
            WHERE attributes ? 'version'
        """
    )
